@@ -11,12 +11,13 @@ use Exporter;
 use Data::Dumper;
 # use Storable qw(dclone);
 use Anarres::Mud::Driver::Program::Type qw(:types);
-use Anarres::Mud::Driver::Program::Efun qw(%EFUNS);
+use Anarres::Mud::Driver::Program::Efun;
 
 sub CONST		()	{ 1 }
 sub LVALUE		()	{ 2 }
 
 sub TYPECHECKNAME	() { 1 }
+sub TYPECHECKALIAS	() { 2 }
 
 BEGIN {
 	my @FLAGS = qw(CONST LVALUE);
@@ -29,7 +30,7 @@ BEGIN {
 		all		=> \@EXPORT_OK,
 			);
 
-	$DEBUG = 0;
+	$DEBUG = 1;
 	%HINTS = ();
 
 	# Vivify the relevant packages
@@ -97,13 +98,12 @@ BEGIN {
 			package $PACKAGE\::$_;
 			use strict;
 			use vars qw(\@ISA);
+			use Carp qw(:DEFAULT cluck);
 			use Data::Dumper;
 			use Anarres::Mud::Driver::Program::Node qw(:all);
 			use Anarres::Mud::Driver::Program::Type qw(:types);
 			use Anarres::Mud::Driver::Compiler qw(\$PROGRAM);
-			use Anarres::Mud::Driver::Program::Efun qw(%EFUNS);
 			\@ISA = qw(Anarres::Mud::Driver::Program::Node);
-			# *EFUNS = \\\%Anarres::Mud::Driver::Program::Efun::EFUNS;
 		}; die $@ if $@;
 	}
 }
@@ -183,6 +183,9 @@ sub infer {
 #	{
 #		my $type = $self->type;
 #		print "Inferring $self from $type to $newtype\n";
+#		if (ref($type) !~ /::Type$/) {
+#			print "Invalid type in " . $self->dump . "\n";
+#		}
 #	}
 	my $node = $self->type->promote($self, $newtype);
 	# print "Node is $node\n" if $node;
@@ -270,6 +273,7 @@ my %OPTYPETABLE = (
 				[ 'ArrIndex', T_STRING, T_STRING->pointer, T_INTEGER ],# sub()?
 				[ 'ArrIndex', T_MIXED, T_ARRAY, T_INTEGER ],# sub()?
 				[ 'MapIndex', T_MIXED, T_MAPPING, T_STRING ],# sub()?
+				[ 'MapIndex', T_MIXED, T_MAPPING, T_OBJECT ],# sub()?
 				],
 	Range	=> [[ 'StrRange', T_STRING, T_STRING, T_INTEGER, T_INTEGER],
 				[ 'ArrRange', T_ARRAY, T_ARRAY, T_INTEGER, T_INTEGER ],
@@ -322,7 +326,7 @@ sub typecheckblock {
 			print "Cannot typecheck $_ in " . $self->nodetype . "\n";
 			next;
 		}
-		# print "Typecheck child " . $_->nodetype . "\n";
+		# print "Typecheck child " . $_->dump . "\n";
 		$succ++ if $_->typecheck($program, @rest);
 	}
 
@@ -343,7 +347,11 @@ sub typecheck {
 	my ($self, $program, @rest) = @_;
 
 	if ($self->type) {
-		cluck "Node " . $self->nodetype . " already has type\n";
+		# This will happen since we have aliasing in the parse
+		# tree, generated by constructs such as "a -= b" becoming
+		# "a = a - b". See Parse.yp for why this is wrong.
+		warn "Node " . $self->nodetype . " already has type: $self"
+						if $DEBUG & TYPECHECKALIAS;
 		return 1;
 	}
 
@@ -473,6 +481,12 @@ my %OPCODETABLE = (
 	StrLe	=> '(A) le (B)',
 	StrGe	=> '(A) ge (B)',
 
+	ArrEq	=> '(A) == (B)',
+	ArrNe	=> '(A) != (B)',
+
+	MapEq	=> '(A) == (B)',
+	MapNe	=> '(A) != (B)',
+
 	IntOr	=> '(A) | (B)',
 	IntAnd	=> '(A) & (B)',
 	IntXor	=> '(A) ^ (B)',
@@ -509,6 +523,7 @@ my %OPCODETABLE = (
 
 	# We can add extra braces around statement|block tokens
 	# This lot are all strictly cheating anyway! If this works ...
+	StmtExp		=> 'A;',
 	StmtDo		=> 'do { B } while (A);',
 	StmtWhile	=> 'while (A) { B }',
 	StmtFor		=> 'for (A; B; C) D',
@@ -516,6 +531,18 @@ my %OPCODETABLE = (
 	StmtTry		=> 'eval A; if ($@) { my B = $@; C; }',
 												# This uses blocks
 	StmtCatch	=> 'eval A;',					# A MudOS hack
+
+	# StmtCase	=> 'C: # case',					# A label (unused?)
+	# StmtDefault	=> 'A: # default',				# A label!
+
+	# This NOGEN business is really developer support and can be removed
+	map { $_ => 'NOGEN' } qw(
+							Variable
+							Index Range
+							Lsh Rsh
+							Add Sub Mul Div Mod
+							Eq Ne Lt Gt Le Ge Or
+							And Xor),
 		);
 
 sub gensub {
@@ -544,7 +571,14 @@ sub generate {
 
 	my $name = $self->nodetype;
 	# print "Finding code for $name\n";
-	my $code = $OPCODETABLE{$name} or return "GEN($name)";
+	my $code = $OPCODETABLE{$name};
+	return "GEN($name)" unless defined $code;
+
+	# This is mostly for debugging. It can be safely removed.
+	if ($code eq 'NOGEN') {
+		warn "Attempt to generate NOGEN opcode $name";
+		return "GEN($name)";
+	}
 
 	my $subref = $self->gensub($name, $code);
 
@@ -623,6 +657,47 @@ sub t_string {	# Modifies its arguments
 	sub typecheck { $_[0]->settype(T_NIL); 1; }
 }
 
+if (0) {
+my %assertmap = (
+	IntAssert	=> 'T_INTEGER',
+	StrAssert	=> 'T_STRING',
+	ArrAssert	=> 'T_ARRAY',
+	MapAssert	=> 'T_MAPPING',
+	ObjAssert	=> 'T_OBJECT',
+		);
+foreach (qw(IntAssert StrAssert ArrAssert MapAssert
+				ObjAssert)) {	# XXX Do ClsAssert
+	my $type = $assertmap{$_};
+	eval qq[
+		{
+			package Anarres::Mud::Driver::Program::Node::$_;
+	] . q[
+			sub typecheck {
+				my ($self, $program, @rest) = @_;
+				my $val = $self->value(0);
+				while ((ref($val) =~ /::...Assert/)) {
+					if (ref($val) eq ref($self)) {
+						$val = $val->value(0);
+						$self->setvalue(0, $val);
+					}
+					else {
+						die "Cannot assert " . ref($val) . " into " .
+										ref($self);
+					}
+				}
+				unless ($val->type) {
+					$val->typecheck($program, @rest)
+									or return undef;
+				}
+	] . qq[
+				\$self->settype($type);
+				return 1;
+			}
+		}
+	]; die $@ if $@;
+}	# foreach
+}	# lexical scope
+
 {
 	package Anarres::Mud::Driver::Program::Node::Nil;
 	sub typecheck { $_[0]->settype(T_NIL); $_[0]->setflag(CONST); 1; }
@@ -775,13 +850,23 @@ sub t_string {	# Modifies its arguments
 	# XXX Look up type
 	sub typecheck {
 		my ($self, $program, @rest) = @_;
+		# cluck "Typechecking Variable $self";
 		my $name = $self->value(0);
-		my $var = $program->variable($name);
-		unless ($var) {
+		my ($var, $class);
+		if ($var = $program->local($name)) {
+			$class = 'Anarres::Mud::Driver::Program::Node::VarLocal';
+		}
+		elsif ($var = $program->global($name)) {
+			$class = 'Anarres::Mud::Driver::Program::Node::VarGlobal';
+		}
+		else {
 			$program->error("Variable $name not found");
+			# We should fake something up because we end up dying later?
 			return undef;
 		}
+		bless $self, $class;
 		$self->settype($var->type);
+		# print "Created variable: " . $self->dump . "\n";
 		return 1;
 	}
 	# XXX As an rvalue? Delegate to a basic type infer method.
@@ -795,16 +880,37 @@ sub t_string {	# Modifies its arguments
 		return "(" . $self->dumptype . "variable "
 						. $self->value(0) . ")";
 	}
+}
+
+{
+	package Anarres::Mud::Driver::Program::Node::VarLocal;
+	sub dump {
+		"(" . $_[0]->dumptype . "varlocal " . $_[0]->value(0) . ")";
+	}
 	sub generate {
-		# XXX This is WRONG if it's a global!
 		return '$' . $_[0]->value(0);
 	}
 }
 
+{
+	package Anarres::Mud::Driver::Program::Node::VarGlobal;
+	sub dump {
+		"(" . $_[0]->dumptype . "varglobal " . $_[0]->value(0) . ")";
+	}
+	sub generate {
+		my $self = shift;
+		my $name = $self->value(0);
+		return '$self->{Variables}->{' . $name . '}';
+	}
+}
+
 {
 	package Anarres::Mud::Driver::Program::Node::Parameter;
-	# XXX Look up type
-	sub old_new { shift; return __PACKAGE__->SUPER::new(T_MIXED, LVALUE, @_); }
+	sub dump {
+		my $self = shift;
+		return "(" . $self->dumptype . "parameter "
+						. $self->value(0) . ")";
+	}
 	# XXX We could look this up at the current point ...
 	sub infer { $_[0]->assert($_[1]) }	# XXX As an rvalue?
 	sub typecheck { $_[0]->settype(T_MIXED); return 1; }
@@ -816,6 +922,8 @@ sub t_string {	# Modifies its arguments
 	# XXX Look up return type, number of args
 	sub typecheck {
 		my ($self, $program, @rest) = @_;
+		# Changing the format of this node will require modifications
+		# to StmtIf optimisation.
 		my @values = $self->values;
 		my $method = shift @values;
 		my @args = @{ $method->args };
@@ -847,25 +955,20 @@ sub t_string {	# Modifies its arguments
 			return undef;
 		}
 
-		my $i = 0;
-		foreach (@args) {
-			my $val = $values[$i];
-			$i++;	# for printing already
-			# print "Matching arg against " . $_->dump . "\n";
-			my $arg = $val->infer($_->type);
-			unless ($arg) {
-				$program->error("Argument $i to " . $method->name .
-								" is type " . ${ $val->type } .
-								" not type " . ${ $_->type });
-				next;
-			}
-			# print "OK\n";
-		}
+		unshift(@values, $method);
+		my $type = $method->typecheck_call($program, \@values);
+
+		return undef unless $type;
 
 #		print "Funcall " . $method->name . " checked and becomes type "
-#						. ${$method->type} . "\n" if 0;
-		$self->settype($method->type);
+#						. ${$type} . "\n" if 0;
 
+		# We overwrite the method object in the first place too.
+		my $i = 0;
+		foreach (@values) {
+			$self->setvalue($i++, $_);
+		}
+		$self->settype($type);
 		return 1;
 	}
 
@@ -882,7 +985,7 @@ sub t_string {	# Modifies its arguments
 		my @args = $self->values;
 		my $method = shift @args;
 		@args = map { $_->generate(@_) } @args;
-		return $method->gencall(@args);
+		return $method->generate_call(@args);
 	}
 }
 
@@ -892,6 +995,8 @@ sub t_string {	# Modifies its arguments
 	sub typecheck {
 		my ($self, $program, @rest) = @_;
 		my ($exp, $name, @values) = $self->values;
+		print "Typechecking CallOther\n"
+					if $DEBUG & TYPECHECKNAME;
 		unshift(@values, $exp);
 		$self->typecheckblock(\@values, $program, @rest)
 						or return undef;
@@ -957,7 +1062,8 @@ sub t_string {	# Modifies its arguments
 		# What we really mean here is, "Is it pure?"
 		# But that would not necessarily amount to an optimisation.
 		# A better question might be, "Is it elementary?"
-		if (ref($val) =~ /::Variable$/ || ($val->flags) & CONST) {
+		# (VarLocal or VarGlobal)
+		if (ref($val) =~ /::Var(Loc|Glob)al$/ || ($val->flags)&CONST) {
 			return $self->generate_cst(@_);
 		}
 		else {
@@ -1087,88 +1193,20 @@ foreach ( qw(Lsh Rsh Mul Div Mod)) {
 	eval qq{
 		{
 			package Anarres::Mud::Driver::Program::Node::$_;
-			sub old_new { shift;
-				t_integer(0, "integer " . lc "$_", \@_);
-				return __PACKAGE__->SUPER::new(T_INTEGER, 0, \@_);
-			}
 		}
 	}; die $@ if $@;
 }
 
 {
 	package Anarres::Mud::Driver::Program::Node::StrAdd;
-	sub old_new { shift;
-		t_string(0, "concatenation", @_);
-		return __PACKAGE__->SUPER::new(T_STRING, 0, @_);
-	}
 }
 
 {
 	package Anarres::Mud::Driver::Program::Node::Add;
-	sub old_new {
-		my (undef, $left, $right) = @_;
-
-		my $li = $left->infer(T_INTEGER);
-		my $ri = $right->infer(T_INTEGER);
-		if ($li && $ri) {
-			# XXX Pass const flags
-			return __PACKAGE__->SUPER::new(T_INTEGER, 0, $li, $ri);
-		}
-
-		my $ls = $left->infer(T_STRING);
-		my $rs = $right->infer(T_STRING);
-		if ($ls && $rs) {
-			return new Anarres::Mud::Driver::Program::Node::Dot(
-							$ls, $rs,
-								);
-		}
-
-		my $la = $left->infer(T_ARRAY);
-		my $ra = $right->infer(T_ARRAY);
-		if ($la && $ra) {
-			return new Anarres::Mud::Driver::Program::Node::ArrAdd(
-							$la, $ra,
-								);
-		}
-
-		my $lm = $left->infer(T_MAPPING);
-		my $rm = $right->infer(T_MAPPING);
-		if ($lm && $rm) {
-			return new Anarres::Mud::Driver::Program::Node::MapAdd(
-							$lm, $rm,
-								);
-		}
-
-		error("Cannot add types " . $left->type->dump . " and " .
-						$right->type->dump);
-	}
 }
 
 {
 	package Anarres::Mud::Driver::Program::Node::Sub;
-	sub old_new {
-		my (undef, $left, $right) = @_;
-
-		my $li = $left->infer(T_INTEGER);
-		my $ri = $right->infer(T_INTEGER);
-		if ($li && $ri) {
-			# XXX Pass const flags
-			return __PACKAGE__->SUPER::new(T_INTEGER, 0, $li, $ri);
-		}
-
-		# No string case
-
-		my $la = $left->infer(T_ARRAY);
-		my $ra = $left->infer(T_ARRAY);
-		if ($la && $ra) {
-			return new Anarres::Mud::Driver::Program::Node::ArrSub(
-							$la, $ra,
-								);
-		}
-
-		error("Cannot subtract types " . $left->type->dump . " and " .
-						$right->type->dump);
-	}
 }
 
 {
@@ -1185,8 +1223,7 @@ foreach ( qw(Lsh Rsh Mul Div Mod)) {
 
 {
 	package Anarres::Mud::Driver::Program::Node::ArrSub;
-	sub old_new { shift; return __PACKAGE__->SUPER::new(T_ARRAY, 0, @_); }
-	# sub infer { $_[1]->arrayp ? $_[0] : undef }
+	# None of these three methods are used.
 	sub invoke {	# Can't really do this inline?
 		my %table = map { $_ => 1 } @{ $_[1] };
 		return [ grep { ! $table{$_} } @{ $_[0] } ]
@@ -1210,7 +1247,7 @@ foreach ( qw(Lsh Rsh Mul Div Mod)) {
 
 {
 	package Anarres::Mud::Driver::Program::Node::ArrUnion;
-	# sub infer { $_[1]->arrayp ? $_[0] : undef }
+	# XXX Generate this inline like ArrSub.
 	sub invoke {
 		my @left = @{ $_[0] };
 		my %table = map { $_ => 1 } @left;
@@ -1229,6 +1266,7 @@ foreach ( qw(Lsh Rsh Mul Div Mod)) {
 
 {
 	package Anarres::Mud::Driver::Program::Node::ArrIsect;
+	# XXX Generate this inline like ArrSub.
 	# sub infer { $_[1]->arrayp ? $_[0] : undef }
 	sub invoke {
 		my @out = ();
@@ -1252,52 +1290,25 @@ foreach (qw(IntEq IntNe IntLt IntGt IntLe IntGe)) {
 	eval qq{
 		{
 			package Anarres::Mud::Driver::Program::Node::$_;
-			sub old_new {
-				my (undef, \$left, \$right) = \@_;
-
-				if (\$left->type->equals(T_STRING) ||
-					\$right->type->equals(T_STRING)) {
-					t_string(0, "string " . lc "$_", \$left, \$right);
-					return new
-							Anarres::Mud::Driver::Program::Node::$strop(
-									T_STRING, 0, \$left, \$right,
-										);
-				}
-				# This can serve for anything except strings.
-				return __PACKAGE__->SUPER::new(
-								T_INTEGER, 0, \$left, \$right,
-									);
-			}
 		}
 		{
 			package Anarres::Mud::Driver::Program::Node::$strop;
-			# new called correctly from above hence no override needed
 		}
 	}; die $@ if $@;
 }
 
 {
 	package Anarres::Mud::Driver::Program::Node::LogOr;
-	sub old_new { shift;
-		return __PACKAGE__->SUPER::new(T_INTEGER, 0, @_);
-	}
 }
 
 {
 	package Anarres::Mud::Driver::Program::Node::LogAnd;
-	sub old_new { shift;
-		return __PACKAGE__->SUPER::new(T_INTEGER, 0, @_);
-	}
 }
 
 foreach (qw(Or And Xor)) {
 	eval qq{
 		{
 			package Anarres::Mud::Driver::Program::Node::$_;
-			sub old_new { shift;
-				t_integer(0, "integer " . lc "$_", \@_);
-				return __PACKAGE__->SUPER::new(T_INTEGER, 0, \@_);
-			}
 		}
 	}; die $@ if $@;
 }
@@ -1310,13 +1321,13 @@ foreach (qw(Or And Xor)) {
 	package Anarres::Mud::Driver::Program::Node::Assign;
 	sub typecheck {
 		my ($self, $program, $flags, @rest) = @_;
-		my ($var, $exp) = $self->values;
+		my ($lval, $exp) = $self->values;
 
 		print "Typechecking Assign\n"
 					if $DEBUG & TYPECHECKNAME;
 
 		# XXX Work more on lvalues ...
-		unless ($var->typecheck($program, $flags | LVALUE, @rest)) {
+		unless ($lval->typecheck($program, $flags | LVALUE, @rest)) {
 			$program->error("Failed to typecheck lvalue to assign");
 			return undef;
 		}
@@ -1325,38 +1336,28 @@ foreach (qw(Or And Xor)) {
 			return undef;
 		}
 
-		my $vtype = $var->type;
+		my $vtype = $lval->type;
 		if ($vtype->equals(T_MIXED)) {
 			$self->settype($exp->type);
 		}
 		else {
 			my $val = $exp->infer($vtype);
 			unless ($val) {
+				my $dump = $lval->dump;
+				$dump =~ s/\s+/ /g;
 				$program->error("Cannot assign type " .
-								$exp->type->dump .
-								" to a variable of type " .
-								$var->type->dump);
+								$exp->type->dump . " to lvalue " .
+								$dump ." of type ". $lval->type->dump);
 				return undef;
 			}
 			$val->typecheck($program, $flags, @rest) unless $val->type;
 			$self->setvalue(1, $val);
 
-			$self->settype($var->type);
+			$self->settype($lval->type);
 		}
 
 		return 1;
 	}
-	sub old_new {
-		my (undef, $var, $exp) = @_;
-		# XXX Check types
-		# Set initialised flag
-		my $val = $exp->infer($var->type);
-		die
-			"Type mismatch in assignment: " .
-				$var->type->dump . " = " . $exp->type->dump
-					unless $val;
-		return __PACKAGE__->SUPER::new($var->type, 0, $var, $val);
-	}
 }
 
 {
@@ -1367,7 +1368,9 @@ foreach (qw(Or And Xor)) {
 		my @stmts  = @{ $self->value(1) };
 
 		$program->save_locals;
-		$program->locals(\@locals);		# XXX Pass orig ref
+		foreach (@locals) {
+			$program->local($_);
+		}
 		my $succ = 0;
 		foreach (@stmts) {
 			$succ++ if $_->typecheck($program, @rest);
@@ -1407,7 +1410,7 @@ foreach (qw(Or And Xor)) {
 
 {
 	package Anarres::Mud::Driver::Program::Node::StmtNull;
-	sub typecheck { $_[0]->settype(T_NIL); 1; }
+	sub typecheck { $_[0]->settype(T_VOID); 1; }
 }
 
 {
@@ -1422,10 +1425,6 @@ foreach (qw(Or And Xor)) {
 		$_[0]->settype(T_VOID);
 		return 1;
 	}
-	sub generate {
-		my $self = shift;
-		return $self->value(0)->generate(@_) . ';';
-	}
 }
 
 {
@@ -1475,7 +1474,7 @@ foreach (qw(Or And Xor)) {
 		my $labels = $program->restore_labels;
 		$self->setvalue(3, $labels->[0]);	# labels
 		$self->setvalue(4, $labels->[1]);	# default
-		$_[0]->settype(T_VOID);
+		$self->settype(T_VOID);
 		return 1;
 	}
 	sub generate {
@@ -1606,6 +1605,14 @@ foreach (qw(Or And Xor)) {
 		$cond->typecheck($program, @rest)
 						or return undef;
 		# Now we inspect $cond and set hints.
+		if (ref($cond) =~ /::Funcall$/) {
+			my $method = $cond->value(0);
+			my $name = $method->name;
+			# intp, stringp, boolp, objectp, classp, arrayp, mapp
+			if ($name =~ /(?:int|string|bool|object|class|array|map)p/){
+				print "Hinting conditional: Call to $name\n";
+			}
+		}
 		my $vals = defined $else
 				? [ $if, $else, ]
 				: [ $if, ];
@@ -1659,7 +1666,7 @@ foreach (qw(
 	}; die $@ if $@;
 }
 
-if (0) {
+if (1) {
 	my $package = __PACKAGE__;
 	no strict qw(refs);
 	my @nogen;
